@@ -38,7 +38,7 @@ struct OSPV_avarray : OSPV_Generic {
   virtual void CLEAR();
   virtual int FETCHSIZE();
   virtual OSSV *traverse(char *keyish);
-  virtual void XSHARE(int on);
+  virtual void ROSHARE_set(int on);
   virtual double _percent_filled();
 };
 
@@ -74,7 +74,7 @@ struct OSPV_hvarray2 : OSPV_Generic {
   virtual double _percent_filled();
   virtual int FETCHSIZE();
   virtual OSSV *traverse(char *keyish);
-  virtual void XSHARE(int on);
+  virtual void ROSHARE_set(int on);
 };
 
 struct OSPV_hvarray2_cs : OSPV_Cursor {
